﻿#include "freetypewidget.h"
#include "grid.h"
#include <QDebug>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QPainter>

FreeTypeWidget::FreeTypeWidget(QWidget *parent)
    : QWidget(parent)
{
    FT_Init_FreeType(&m_library);

    this->init_widget();
}

FreeTypeWidget::~FreeTypeWidget()
{
    FT_Done_Face(m_face);
    FT_Done_FreeType(m_library);
}

void FreeTypeWidget::init_widget()
{
    QGraphicsScene * scene = new QGraphicsScene(this);
    scene->setItemIndexMethod(QGraphicsScene::NoIndex);

    m_grid = new Grid(this);
    m_grid->set_scene(scene);

    m_graphics_view = new QGraphicsView(this);
    m_graphics_view->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    m_graphics_view->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_graphics_view->setScene(scene);
}

void FreeTypeWidget::set_font(const QString & fontPath, uint font_size)
{
    m_init_finish = false;

    FT_Error res = FT_New_Face(m_library, fontPath.toStdString().c_str(), 0, &m_face);
    if (res)
    {
        qDebug() << "new freetype face failed";
        return;
    }

    res = FT_Set_Pixel_Sizes(m_face, font_size, 0);
    if (res)
    {
        qDebug() << "set font size failed";
        return;
    }

    FT_Select_Charmap(m_face, FT_ENCODING_UNICODE);

    m_init_finish = true;
}

void FreeTypeWidget::set_string(const std::wstring & str, uint manual_height)
{
    if (!m_init_finish) return;

    m_sentence = str;
    m_manual_height = manual_height;

    this->calculate_string_box();

    int width = (m_bbox.xMax - m_bbox.xMin) + 10;
    int height = m_manual_height != 0 ? m_manual_height : (m_bbox.yMax - m_bbox.yMin) + border_top * 2;

    m_graphics_view->scene()->setSceneRect(0, 0, width * m_grid->get_gap() + 100, height * m_grid->get_gap() + 50);
    m_grid->set_pixel_size(width, height).generate_grid();

    emit this->grid_pixel_changed(width, height);
}

void FreeTypeWidget::paintEvent(QPaintEvent * event)
{
    this->render_string(2, m_manual_height != 0 ? 0 : border_top);

    QWidget::paintEvent(event);
}

void FreeTypeWidget::resizeEvent(QResizeEvent *event)
{
    m_graphics_view->resize(this->width(), this->height());

    QWidget::resizeEvent(event);
}

void FreeTypeWidget::draw_bitmap(FT_Bitmap * bitmap, FT_Int x, FT_Int y)
{
    FT_Int i = 0, j = 0, p = 0, q = 0;

    FT_Int x_max = x + bitmap->width;
    FT_Int y_max = y + bitmap->rows;

    for (j = y, q = 0; j < y_max; j++, q++)
    {
        for (i = x, p = 0; i < x_max; i++, p++)
        {
            if (i < 0 || j < 0 || i >= m_grid->get_pixel_width() || j >= m_grid->get_pixel_height())
                continue;

            unsigned char c = bitmap->buffer[q * bitmap->width + p];
            if (c == 0) continue;

            m_grid->update_pixel(i, j, Qt::black);
        }
    }
}

void FreeTypeWidget::calculate_string_box()
{
    m_bbox.xMin = m_bbox.yMin = std::numeric_limits<FT_Pos>::max();
    m_bbox.xMax = m_bbox.yMax = std::numeric_limits<FT_Pos>::min();

    FT_Vector pen = {0, 0};

    FT_GlyphSlot slot = m_face->glyph;

    for (const auto & c : m_sentence)
    {
        FT_Set_Transform(m_face, 0, &pen);

        FT_Error res = FT_Load_Char(m_face, c, FT_LOAD_RENDER);
        if (res)
        {
            qDebug() << "FT_Load_Char error";
            continue;
        }

        FT_Glyph glyph;
        res = FT_Get_Glyph(m_face->glyph, &glyph);
        if (res)
        {
            qDebug() << "FT_Get_Glyph error";
            continue;
        }

        FT_BBox glyph_bbox;
        FT_Glyph_Get_CBox(glyph, FT_GLYPH_BBOX_TRUNCATE, &glyph_bbox);

        m_bbox.xMin = std::min(m_bbox.xMin, glyph_bbox.xMin);
        m_bbox.yMin = std::min(m_bbox.yMin, glyph_bbox.yMin);
        m_bbox.xMax = std::max(m_bbox.xMax, glyph_bbox.xMax);
        m_bbox.yMax = std::max(m_bbox.yMax, glyph_bbox.yMax);

        pen.x += slot->advance.x;
        pen.y += slot->advance.y;
    }

    qDebug() << m_bbox.xMin << m_bbox.xMax << m_bbox.yMin << m_bbox.yMax;
}


void FreeTypeWidget::render_string(int x, int y)
{
    if (m_sentence.empty() || !m_init_finish) return;

    int posx = x;
    int posy = m_grid->get_pixel_width() - y;

    FT_Vector pen = {0, 0};
    FT_GlyphSlot slot = m_face->glyph;

    pen.x = (posx - m_bbox.xMin) * 64;
    pen.y = (posy - m_bbox.yMax) * 64;

    for (const auto & c : m_sentence)
    {
        FT_Set_Transform(m_face, 0, &pen);

        FT_Error res = FT_Load_Char(m_face, c, FT_LOAD_RENDER);
        if (res)
        {
            qDebug() << "FT_Load_Char error";
            continue;
        }

        draw_bitmap(&slot->bitmap, slot->bitmap_left, m_grid->get_pixel_width() - slot->bitmap_top);

        pen.x += slot->advance.x;
        pen.y += slot->advance.y;
    }
}
